﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2019 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Random;
using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Algorithms.EvolvmentModelsOfModels {
  [Item("EMMMultyPointMutator", "Selects a random tree node and changes the symbol.")]
  [StorableType("918B5F77-3B9E-4620-94A3-236913C4A3F2")]
  public sealed class EMMMultyPointsMutator : SymbolicExpressionTreeManipulator {
    private const string ModelSetParameterName = "Models";
    private const string ClusterNumberParameterName = "ClusterNumber";
    private const string MapParameterName = "Map";
    private const string MutationProbabilityParameterName = "MutationProbability";
    public ILookupParameter<ItemList<ISymbolicExpressionTree>> ModelSetParameter {
      get { return (ILookupParameter<ItemList<ISymbolicExpressionTree>>)Parameters[ModelSetParameterName]; }
    }
    public ILookupParameter<ItemList<IntValue>> ClusterNumberParameter {
      get { return (ILookupParameter<ItemList<IntValue>>)Parameters[ClusterNumberParameterName]; }
    }

    public ILookupParameter<ItemList<ItemList<IntValue>>> MapParameter {
      get { return (ILookupParameter<ItemList<ItemList<IntValue>>>)Parameters[MapParameterName]; }
    }
    public ILookupParameter<PercentValue> MutationProbabilityParameter {
      get { return (ILookupParameter<PercentValue>)Parameters[MutationProbabilityParameterName]; }
    }
    public ItemList<ISymbolicExpressionTree> ModelSet => ModelSetParameter.ActualValue;
    public ItemList<IntValue> ClusterNumber => ClusterNumberParameter.ActualValue;
    public ItemList<ItemList<IntValue>> Map => MapParameter.ActualValue;
    public PercentValue MutationProbability => MutationProbabilityParameter.ActualValue;

    [StorableConstructor]
    private EMMMultyPointsMutator(StorableConstructorFlag _) : base(_) { }
    private EMMMultyPointsMutator(EMMMultyPointsMutator original, Cloner cloner) : base(original, cloner) { }
    public EMMMultyPointsMutator() : base() {
      Parameters.Add(new LookupParameter<ItemList<ISymbolicExpressionTree>>(ModelSetParameterName));
      Parameters.Add(new LookupParameter<ItemList<IntValue>>(ClusterNumberParameterName));
      Parameters.Add(new LookupParameter<ItemList<ItemList<IntValue>>>(MapParameterName));
      Parameters.Add(new LookupParameter<PercentValue>(MutationProbabilityParameterName));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new EMMMultyPointsMutator(this, cloner);
    }

    protected override void Manipulate(IRandom random, ISymbolicExpressionTree symbolicExpressionTree) {
      EMMOnePointMutatorPart(random, symbolicExpressionTree, ModelSet, ClusterNumber, Map, MutationProbability);
    }
    private static bool SymbolTypeCheck(ISymbol nSymbol, IEnumerable<ISymbol> allSymbols) {
      foreach (var pSymbol in allSymbols) {
        if (nSymbol == pSymbol)
          return true;
      }
      return false;
    }
    public static void EMMOnePointMutatorPart(IRandom random, ISymbolicExpressionTree symbolicExpressionTree, ItemList<ISymbolicExpressionTree> modelSet, ItemList<IntValue> clusterNumber, ItemList<ItemList<IntValue>> map, PercentValue mutationProbability) {

      List<ISymbol> allowedSymbols = new List<ISymbol>();
      var probability = (1 / mutationProbability.Value) / (symbolicExpressionTree.Length); // probability for edch node to mutate
      var allAllowedSymbols = symbolicExpressionTree.Root.Grammar.AllowedSymbols.Where(s =>
        !(s is Defun) &&
        !(s is GroupSymbol) &&
        !(s is ProgramRootSymbol) &&
        !(s is StartSymbol)
      );

      symbolicExpressionTree.Root.ForEachNodePostfix(node => {//for each node in tree - try to mutate

        if (SymbolTypeCheck(node.Symbol, allAllowedSymbols)) { // we do not want to toch StartSymbol
          if (random.NextDouble() < probability) { // ok. this node decide to mutate

            foreach (var symbol in allAllowedSymbols) {// add in list all nodes with the same arity
              if ((symbol.MaximumArity == node.Symbol.MaximumArity) && (symbol.MinimumArity == node.Symbol.MinimumArity)) { allowedSymbols.Add(symbol); }
            }
            if (node.Symbol.MaximumArity == 0) { // for terminal nodes add aditional the same type of nodes
              allowedSymbols.Add(node.Symbol);
            }
            int pTemp = random.Next(map.Count);
            if (node is TreeModelTreeNode treeNode) { pTemp = treeNode.ClusterNumer; } //remeber the cluster number
            var weights = allowedSymbols.Select(s => s.InitialFrequency).ToList(); // set up probabilities

#pragma warning disable CS0618 // Type or member is obsolete
            var newSymbol = allowedSymbols.SelectRandom(weights, random); // create new node from the choosen symbol type. Ror terminal nodes it means that we have only one possible varint.
#pragma warning restore CS0618 // Type or member is obsolete
            node = newSymbol.CreateTreeNode();

            if (node is TreeModelTreeNode treeNode2) { // make rigth mutation for tree model
              treeNode2.TreeNumber = map[pTemp].SampleRandom(random).Value;
              treeNode2.Tree = (ISymbolicExpressionTree)modelSet[treeNode2.TreeNumber].Clone();
              treeNode2.ClusterNumer = pTemp;
              treeNode2.SetLocalParameters(random, 1);
            } else if (node.HasLocalParameters) { // make local parametrs set up for other node types
              node.ResetLocalParameters(random);
            }
            allowedSymbols.Clear(); // clean before the next node
          }

        }
      });
    }
  }
}